/*
 *   Copyright (c) 2002 Christian W. Zuckschwerdt <zany@triq.net>
 *
 *   This program is free software; you can redistribute it and/or modify it
 *   under the terms of the GNU General Public License as published by the Free
 *   Software Foundation; either version 2 of the License, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful, but
 *   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *   for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>

#include <stdio.h> /* sscanf - replace with xml parser asap */
#include <unistd.h> /* sscanf - replace with xml parser asap */

#include <string.h>
#define _GNU_SOURCE
#include <getopt.h>

#include <obexftp/obexftp.h>
#include <obexftp/client.h>
#include <obexftp/object.h>
#include <cobexbfb/cobex_bfb.h>

gchar *dev = NULL;
obexftp_client_t *cli = NULL;
obex_ctrans_t *ctrans = NULL;
gint input_id = 0;

GSList *object_queue = NULL;
GSList *node_queue = NULL;

GdkPixmap *fo_pix;
GdkPixmap *fc_pix;
GdkPixmap *f_pix;
GdkBitmap *fo_mask;
GdkBitmap *fc_mask;
GdkBitmap *f_mask;

GdkPixmap *busy1_pix;
GdkPixmap *busy2_pix;
GdkPixmap *idle_pix;
GdkBitmap *busy1_mask;
GdkBitmap *busy2_mask;
GdkBitmap *idle_mask;

GtkWidget *statusbar = NULL;
GtkWidget *memorybar = NULL;
GtkCTree *ctree = NULL;
GtkPixmap *busy = NULL;

guint context_id;

guint32 memory_size = 0;
guint32 memory_free = 0;


static void enqueue_request (obex_object_t *object)
{
	/* enqueue then check to avoid race condition */
	object_queue = g_slist_append (object_queue, object);

	if (cli->finished) {
		cli->finished = FALSE;
		OBEX_Request (cli->obexhandle, object);
	}
}


/*
<?xml version="1.0"?>
<!DOCTYPE folder-listing SYSTEM "obex-folder-listing.dtd">
<folder-listing version="1.0">
   <folder name="" modified="20010101T000000" user-perm="WD" group-perm="W" />
   <folder name="Voice memo" modified="" user-perm="RWD" group-perm="W" />
   <folder name="SMS" modified="20011120" user-perm="RWD" group-perm="W" />
   <folder name="Ringing tone" modified="8" user-perm="RWD" group-perm="W" />
   <folder name="Bitmap" modified="2001124" user-perm="RWD" group-perm="W" />
   <folder name="Misc" modified="20048" user-perm="RWD" group-perm="W" />
   <file name="README" size="646" modified="" user-perm="RWD" group-perm="R" />
   <folder name="telecom" />
</folder-listing>
*/

/* FIXME: don't depend on list being \0 terminated. */
static gint parse_list (const gchar *list, gint length)
{
	gchar **lines;
	gchar **p;
	gchar name[200]; // bad coder
	gchar mod[200]; // - no biscuits!
	gchar size[200]; // gint would be ok too.
	gchar *entry[] = {name, size, mod};
	GtkCTreeNode *node;
	GtkCTreeNode *parent;

	gint year;
	gint day;
	gint month;
	gint hour;
	gint min;
	gint sec;

	list = g_strndup (list, length); /* hehe */

	if (node_queue) {
		parent = node_queue->data;
		node_queue = g_slist_remove (node_queue,
					     node_queue->data);
	} else {
		parent = NULL;
	}

	lines = g_strsplit (list, "\n", 0);

	for (p = lines; *p != NULL; p++) {
		g_strstrip (*p);
		if (2 == sscanf (*p, "<folder name=\"%[^\"]\" modified=\"%[^\"]\"", name, mod)) {
			*size = '\0';

			if (6 == sscanf(mod, "%4d%2d%2dT%2d%2d%2d", &year,
					&day, &month, &hour, &min, &sec)) {
				g_snprintf (mod, 200, 
					    "%4d-%02d-%02d %2d:%02d:%02d",
					    year, month, day, hour, min, sec);
			}
			
			node = gtk_ctree_insert_node (ctree,
						      parent, NULL,
						      entry, 0,
						      fo_pix, fo_mask,
						      fc_pix, fc_mask,
						      FALSE, TRUE);
			gtk_ctree_node_set_row_data (ctree, node, g_strdup (name));

			// we need a nested path build - this is broken
			if (parent == NULL) {
			object_queue = g_slist_append (object_queue,
			       obexftp_build_list (cli->obexhandle,
						   g_strdup (name)));
			node_queue = g_slist_append (node_queue, node);
			}
		}
		if (3 == sscanf (*p, "<file name=\"%[^\"]\" size=\"%[^\"]\" modified=\"%[^\"]\"", name, size, mod)) {

			if (6 == sscanf(mod, "%4d%2d%2dT%2d%2d%2d", &year,
					&day, &month, &hour, &min, &sec)) {
				g_snprintf (mod, 200, 
					    "%4d-%02d-%02d %2d:%02d:%02d",
					    year, month, day, hour, min, sec);
			}

			node = gtk_ctree_insert_node (ctree,
						      parent, NULL,
						      entry, 0,
						      f_pix, f_mask,
						      NULL, NULL,
						      TRUE, TRUE);
			gtk_ctree_node_set_row_data (ctree, node, g_strdup (name));
		}
		// handle hidden folder!

	}

	g_strfreev (lines);

	return 0;
}


void info_cb (gint event, const gchar *param, gint len, gpointer data)
{
	switch (event) {

	case OBEXFTP_EV_ERRMSG:
		g_print ("Error: %s\n", param);
		break;

	case OBEXFTP_EV_ERR:
		g_print ("failed: %s\n", param);
		break;
	case OBEXFTP_EV_OK:
		g_print ("done\n");
		break;

	case OBEXFTP_EV_CONNECTING:
		g_print ("Connecting...");
		break;
	case OBEXFTP_EV_DISCONNECTING:
		g_print ("Disconnecting...");
		break;
	case OBEXFTP_EV_SENDING:
		g_print ("Sending %s...", param);
		break;
	case OBEXFTP_EV_RECEIVING:
		g_print ("Receiving %s...", param);
		break;

	case OBEXFTP_EV_LISTENING:
		g_print ("Waiting for incoming connection\n");
		break;

	case OBEXFTP_EV_CONNECTIND:
		g_print ("Incoming connection\n");
		break;
	case OBEXFTP_EV_DISCONNECTIND:
		g_print ("Disconnecting\n");
		break;

	case OBEXFTP_EV_BODY:
		parse_list (param, len);
		break;

	case OBEXFTP_EV_INFO:
		g_print ("Got Info: %d\n", GPOINTER_TO_UINT (param));
		if (memory_size == 0)
			memory_size = GPOINTER_TO_UINT (param);
		else if (memory_free == 0)
			memory_free = GPOINTER_TO_UINT (param);

		gtk_progress_configure (GTK_PROGRESS (memorybar),
					memory_free, 0, memory_size);

		break;

	}
}


static void input_function (gpointer data, gint source,
			      GdkInputCondition condition)
{
	static gint busy_count = 0;
	OBEX_HandleInput (cli->obexhandle, 0);
	g_print ("handle input done: %s\n", cli->finished ? "finished" : "working");

	/* toggle busy indicator*/
	if (cli->finished)
		gtk_pixmap_set (busy, idle_pix, idle_mask);
	else if (busy_count++ % 2 == 0)
		gtk_pixmap_set (busy, busy1_pix, busy1_mask);
	else
		gtk_pixmap_set (busy, busy2_pix, busy2_mask);


	/* process the next object if the link is ready */
	if (cli->finished) {
		if (object_queue) {
			cli->finished = FALSE;
			OBEX_Request (cli->obexhandle, object_queue->data);
			object_queue = g_slist_remove (object_queue,
						       object_queue->data);
		}
	}
}


void gobexftp_set_icons (GtkPixmap *folder_opened_icon,
			 GtkPixmap *folder_closed_icon,
			 GtkPixmap *file_icon)
{
	gtk_pixmap_get (folder_opened_icon, &fo_pix, &fo_mask);
	gtk_pixmap_get (folder_closed_icon, &fc_pix, &fc_mask);
	gtk_pixmap_get (file_icon, &f_pix, &f_mask);
}


void gobexftp_set_pixmaps (GtkPixmap *busy1_pixmap,
			   GtkPixmap *busy2_pixmap,
			   GtkPixmap *idle_pixmap)
{
	gtk_pixmap_get (busy1_pixmap, &busy1_pix, &busy1_mask);
	gtk_pixmap_get (busy2_pixmap, &busy2_pix, &busy2_mask);
	gtk_pixmap_get (idle_pixmap, &idle_pix, &idle_mask);
}


void gobexftp_set_device (gchar *device)
{
	if (!g_strcasecmp("irda", device))
		dev = NULL;
	else
		dev = device;
}


void gobexftp_set_statusbar (GtkWidget *widget)
{
	statusbar = widget;
	context_id = gtk_statusbar_get_context_id (GTK_STATUSBAR (statusbar),
						   "GObexFtp");
}


void gobexftp_set_memorybar (GtkWidget *widget)
{
	memorybar = widget;
}


void gobexftp_set_ctree (GtkWidget *widget)
{
	ctree = GTK_CTREE (widget);
}


void gobexftp_set_busy (GtkWidget *widget)
{
	busy = GTK_PIXMAP (widget);
}


gint gobexftp_connect (void)
{
	if(cli != NULL) {
		gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
				    "Already connected!");
		return 0;
	}

	gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
			    "Connecting...");

	g_print("Connecting to %s\n", dev);

	if (dev != NULL)
		ctrans = cobex_ctrans (dev);
	else
		ctrans = NULL;

	cli = obexftp_cli_open(info_cb, ctrans, NULL);
	if(cli == NULL) {
		g_print("Error opening obexftp-client\n");
		return -1;
	}

	/* Connect */
	if(obexftp_cli_connect(cli) < 0) {
		obexftp_cli_close(cli);
		cli = NULL;
		gtk_statusbar_pop (GTK_STATUSBAR (statusbar), context_id);
		gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
				    "Error connecting!");
		return -1;
	}

	/* register OBEX input handler */
	input_id = gdk_input_add(OBEX_GetFD(cli->obexhandle),
				 GDK_INPUT_READ,
				 &input_function,
				 cli);

	gtk_statusbar_pop (GTK_STATUSBAR (statusbar), context_id);
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
			    "Connected!");

	return 0;
}


gint gobexftp_disconnect (void)
{
	if(cli == NULL) {
		gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
				    "Not connected!");
		return -1;
	}

	gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
			    "Disconnecting...");

	/* clear queue */
	while (object_queue) {
		object_queue = g_slist_remove (object_queue,
					       object_queue->data);
	}

	/* unregister OBEX input handler */
	gdk_input_remove(input_id);

	/* Disconnect */
	obexftp_cli_disconnect(cli);

	obexftp_cli_close(cli);
	cobex_free (ctrans);
	cli = NULL;

	gtk_statusbar_pop (GTK_STATUSBAR (statusbar), context_id);
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
			    "Disconnected!");

	return 0;
}


gint gobexftp_info (void)
{
	if(cli == NULL) {
		gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
				    "Not connected!");
		return -1;
	}

	// Retrieve Info
	cli->finished = FALSE;
	OBEX_Request (cli->obexhandle,
		      obexftp_build_info (cli->obexhandle, 0x01));
	object_queue = g_slist_append (object_queue,
		      obexftp_build_info (cli->obexhandle, 0x02));

	return 0;
}


gint gobexftp_list_node (GtkCTreeNode *node)
{
	gchar *text = NULL;
	gboolean leaf;

	if(cli == NULL) {
		gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id,
				    "Not connected!");
		return -1;
	}

	// Get folder name
	if (node != NULL)
		gtk_ctree_get_node_info (ctree, node, &text,
					 NULL, NULL, NULL, NULL, NULL,
					 &leaf, NULL);
		
	// Retrieve listing
	enqueue_request (obexftp_build_list (cli->obexhandle, NULL));

	/*
	for (; node != NULL; node = GTK_CTREE_NODE_NEXT (node)) {
		gtk_ctree_get_node_info (ctree, node, &text,
					 NULL, NULL, NULL, NULL, NULL,
					 &leaf, NULL);
		if (!leaf)
			g_print ("Item: %s\n", text);
	}

	*/

	return 0;
}


gint gobexftp_list (void)
{
	gobexftp_list_node (gtk_ctree_node_nth (ctree, 0));

	return 0;
}


void gobexftp_clear (void)
{
	GtkCTreeNode *node;

	/* is this how it's supposed to be done? */
	while ((node = gtk_ctree_node_nth (ctree, 0)) != NULL)
		gtk_ctree_remove_node (ctree, node);
}


static gchar *selected_node_data (void)
{
	GList *sel = GTK_CLIST (ctree)->selection;
	if (sel == NULL)
		/* should not happen */
		return NULL;

	return gtk_ctree_node_get_row_data (ctree, GTK_CTREE_NODE (sel->data));
}


void gobexftp_upload (gchar *fn)
{
	GList *sel;
	gchar *text;

	sel = GTK_CLIST (ctree)->selection;
	if (sel == NULL)
		/* should not happen */
		return;

	gtk_ctree_node_get_text (ctree, GTK_CTREE_NODE (sel->data), 0, &text);
	printf ("1 %s -> %s\n", text, fn);
	gtk_ctree_node_get_text (ctree, GTK_CTREE_NODE (sel->data), 1, &text);
	printf ("2 %s -> %s\n", text, fn);
	gtk_ctree_node_get_text (ctree, GTK_CTREE_NODE (sel->data), 2, &text);
	printf ("3 %s -> %s\n", text, fn);
	text = gtk_ctree_node_get_row_data (ctree, GTK_CTREE_NODE (sel->data));
	printf ("d %s -> %s\n", text, fn);

	if (!cli->finished);
		return;
/*
	cli->finished = FALSE;
	OBEX_Request (cli->obexhandle,
		      build_object_from_file (cli->obexhandle,
					      fn, / *remote* /fn));
*/
}


void gobexftp_download (gchar *fn)
{
	printf ("obex_download %s -> %s\n", selected_node_data (), fn);

	if (!cli->finished);
		return;
	cli->finished = FALSE;
	cli->target_fn = fn;
	OBEX_Request (cli->obexhandle,
		      obexftp_build_get (cli->obexhandle,
					 selected_node_data ()));
}


void gobexftp_rename (gchar *fn)
{
	printf ("obex_rename %s -> %s\n", selected_node_data (), fn);

	if (!cli->finished);
		return;
	cli->finished = FALSE;
	OBEX_Request (cli->obexhandle,
		      obexftp_build_rename (cli->obexhandle,
					    selected_node_data (),
					    fn));
}


void gobexftp_delete ()
{
	printf ("obex_delete %s\n", selected_node_data ());

	if (!cli->finished)
		return;
	cli->finished = FALSE;
	OBEX_Request (cli->obexhandle,
		      obexftp_build_del (cli->obexhandle,
					 selected_node_data ()));
}

